/*
	muhaqwi.c	[mu' HaqwI' - word surgeon]
	Klingon analyzer and dictionary

	(c) 1995-96 by Christoph "bIQHurgh" Pfisterer <chrisp@ardanwen.sax.de>
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*** SWITCHES ***/

#define	MAX_IDX	2000
#define	MAX_ANUM	30
#define	STRING_LEN	64

/*** DEFINES ***/

#define	FALSE	0
#define	TRUE	1

#define	FL_NOUN  (1 << 0)
#define	FL_NSUF  (1 << 1)
#define	FL_NAME  (1 << 2)
#define	FL_VERB  (1 << 3)
#define	FL_VPRE  (1 << 4)
#define	FL_VSUF  (1 << 5)
#define	FL_PRO   (1 << 6)
#define	FL_NUM   (1 << 7)
#define	FL_NUSUF (1 << 8)
#define	FL_OTHER (1 << 9)
#define	FL_MARK  (1 << 13)

/*** TYPES ***/

typedef struct {
	char	tc[4];
	char	s[1];
} VOCAB_REC;

typedef struct {
	short	fl, inum, idx[20];
	char	s[STRING_LEN];
} ANAL_REC;

/*** VARIABLES ***/

char		input[256];
VOCAB_REC	*vl[MAX_IDX];
short		vc[MAX_IDX];
char		vf[MAX_IDX];
short		vnum = 0, anum, mode = 0;
ANAL_REC	al[MAX_ANUM];
short		cont = TRUE;

char	alph[] = "abccDeggHIjlmnnopqQrSttuvwy'";

unsigned char upcase[256] = {
 0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,0x0f,
 0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f,
 0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,0x29,0x2a,0x2b,0x2c,0x2d,0x2e,0x2f,
 0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,0x3e,0x3f,
 0x40,0x41,0x42,0x43,0x44,0x45,0x46,0x47,0x48,0x49,0x4a,0x4b,0x4c,0x4d,0x4e,0x4f,
 0x50,0x51,0x52,0x53,0x54,0x55,0x56,0x57,0x58,0x59,0x5a,0x5b,0x5c,0x5d,0x5e,0x5f,
 0x60,0x41,0x42,0x43,0x44,0x45,0x46,0x47,0x48,0x49,0x4a,0x4b,0x4c,0x4d,0x4e,0x4f,
 0x50,0x51,0x52,0x53,0x54,0x55,0x56,0x57,0x58,0x59,0x5a,0x7b,0x7c,0x7d,0x7e,0x7f,
 0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8a,0x8b,0x8c,0x8d,0x8e,0x8f,
 0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9a,0x9b,0x9c,0x9d,0x9e,0x9f,
 0xa0,0xa1,0xa2,0xa3,0xa4,0xa5,0xa6,0xa7,0xa8,0xa9,0xaa,0xab,0xac,0xad,0xae,0xaf,
 0xb0,0xb1,0xb2,0xb3,0xb4,0xb5,0xb6,0xb7,0xb8,0xb9,0xba,0xbb,0xbc,0xbd,0xbe,0xbf,
 0xc0,0xc1,0xc2,0xc3,0xc4,0xc5,0xc6,0xc7,0xc8,0xc9,0xca,0xcb,0xcc,0xcd,0xce,0xcf,
 0xd0,0xd1,0xd2,0xd3,0xd4,0xd5,0xd6,0xd7,0xd8,0xd9,0xda,0xdb,0xdc,0xdd,0xde,0xdf,
 0xe0,0xe1,0xe2,0xe3,0xe4,0xe5,0xe6,0xe7,0xe8,0xe9,0xea,0xeb,0xec,0xed,0xee,0xef,
 0xf0,0xf1,0xf2,0xf3,0xf4,0xf5,0xf6,0xf7,0xf8,0xf9,0xfa,0xfb,0xfc,0xfd,0xfe,0xff,
};

/*** FUNCTIONS ***/

void show_help (void);
void process_cmd (char *input);

void analyzer (char *input);
void anal_word (char *s);
short anal_match (char *s, char *m);

void dictionary (char *input);
short dict_match (char *s, char *pat);

void read_mutay (void);
int voc_sort (const void *v1, const void *v2);

/*
**	main
*/

int main (void)
{
	puts ("mu' HaqwI' 1.0 - enter '?' for help");
	read_mutay ();

	while (cont)
	{
		fputs ("SuH> ", stdout);
		gets (input);
		if (input[0] == 0)
			cont = FALSE;
		else if (input[0] == '?')
			show_help ();
		else if (input[0] == '!')
			process_cmd (input+1);
		else if (mode == 0)
			analyzer (input);
		else
			dictionary (input);
	}

	puts ("Qapla'");
	return 0;
}

/*
**	show help message
*/

void show_help (void)
{
	puts ("mu' HaqwI' [word surgeon] version 1.0");
	puts ("(c) 1995-96 by Christoph \"bIQHurgh\" Pfisterer");
	puts ("");
	puts ("Commands consist of a single letter preceded by an exclamation mark.");
	puts ("Everything not flagged this way is sent to the analyzer or the dictionary,");
	puts ("depending on the current mode. An empty line will quit the program.");
	puts ("");
	puts ("The analyzer will parse your input looking for valid Klingon words. You may");
	puts ("enter multiple words on one line, separated by spaces or tabs. The analyzer");
	puts ("treats punctuation characters (e.g. .,:;!? ) like spaces. It will fit your");
	puts ("input to the Klingon alphabet before analyzing it (e.g. BoRTas -> bortaS,");
	puts ("gic -> ghIch).");
	puts ("");
	puts ("The dictionary will use your input as a pattern string to look for. Again, you");
	puts ("may enter multiple words separated by spaces or tabs, but they are used");
	puts ("literally. The dictionary will display any word that contains at least one of");
	puts ("the pattern strings, either in the Klingon word itself or in the English");
	puts ("translation. The search is case-insensitive.");
	puts ("");
	puts ("Currently available commands are:");
	puts ("  !A   switch to analyzer mode");
	puts ("  !D   switch to dictionary mode");
	puts ("  !Q   quit mu' HaqwI'");
}

/*
**	process a command
*/

void process_cmd (char *input)
{
	switch (upcase[(unsigned char)*input])
	{
	case 'A':
		puts ("Analyzer mode set");
		mode = 0;
		break;
	case 'D':
		puts ("Dictionary mode set");
		mode = 1;
		break;
	case 'Q':
		cont = FALSE;
		break;
	default:
		puts ("Unknown command");
		break;
	}
}

/*
**	entry point for the analyzer
*/

void analyzer (char *input)
{
	short	i, j, wcnt;
	char	c, *p, *d, word[STRING_LEN], buf[256];
	short	mark_for_quit = FALSE;

	wcnt = 0;

	p = input;
	d = buf;
	while ((c = *p++) != 0)
	{
		switch (c)
		{
		case 'A':
		case 'B':
		case 'E':
		case 'J':
		case 'L':
		case 'M':
		case 'O':
		case 'P':
		case 'R':
		case 'U':
		case 'V':
		case 'W':
		case 'Y':
			c += 32;
			break;
		case 'd':
		case 'h':
		case 'i':
		case 's':
			c -= 32;
			break;
		case 'C':
		case 'c':
			if (*p == 'h' || *p == 'H')
				p++;
			*d++ = 'c';
			c = 'h';
			break;
		case 'G':
		case 'g':
			if (*p == 'h' || *p == 'H')
				p++;
			*d++ = 'g';
			c = 'h';
			break;
		case 'N':
		case 'n':
			if (*p == 'g' || *p == 'G')
			{
				p++;
				*d++ = 'n';
				c = 'g';
			}
			else
				c = 'n';
			break;
		case 'T':
		case 't':
			if (*p == 'l' || *p == 'L')
			{
				p++;
				if (*p == 'h' || *p == 'H')
				{
					p++;
					*d++ = 't';
					*d++ = 'l';
					c = 'h';
				}
				else
				{
					*d++ = 't';
					c = 'l';
				}
			}
			else
				c = 't';
			break;
		case '`':
			c = '\'';
			break;
		case '.':
		case ',':
		case ';':
		case ':':
		case '!':
		case '?':
			c = ' ';
			break;
		}
		*d++ = c;
	}
	*d = 0;
	strcpy (input, buf);

	p = input;
	while (!mark_for_quit)
	{
		for (; (c = *p) != 0; p++)
			if (c != 32 && c != 9 && c != 13)
				break;
		if (c == 0)
			break;

		for (d = word; (c = *p) != 0; p++)
		{
			if (c == 32 || c == 9 || c == 13)
				break;
			*d++ = c;
		}
		if (c == 0)
			mark_for_quit = TRUE;
		*d = 0;

		anal_word (word);

		if (wcnt)
			puts ("=====");
		if (anum)
		{
			for (i = 0; i < anum; i++)
			{
				if (i > 0)
					puts ("+----");
				for (j = al[i].inum; --j >= 0;)
				{
					char	*p;

					strcpy (buf, "| ");
					p = vl[al[i].idx[j]]->s;
					strcat (buf, p);
					p = strchr (p, 0) + 1;
					while (strlen (buf) < 12)
						strcat (buf, " ");
					strcat (buf, " ???? ");
					memcpy (strchr (buf, 0) - 5, vl[al[i].idx[j]]->tc, 4L);
					strcat (buf, p);
					puts (buf);
				}
			}
		}
		else
		{
			strcpy (buf, "| ");
			strcat (buf, word);
			while (strlen (buf) < 12)
				strcat (buf, " ");
			strcat (buf, " ?    unknown");
			puts (buf);
		}
		wcnt++;
	}
}

/*
**	the real analysis
*/

void anal_word (char *s)
{
	short	i, j, ai, sl, nf, af;
	char	real_as[STRING_LEN+2], *as = real_as + 2, *asend;

	if (*s == 0)
	{
		anum = 0;
		return;
	}

	anum = 1;
	al[0].fl = 1007;
	al[0].inum = 0;
	strcpy (al[0].s, s);
	as[-1] = 0;

	for (;;)
	{
		as[0] = 0;
		for (ai = 0; ai < anum; ai++)
			if (strlen(al[ai].s) > strlen(as))
				strcpy (as, al[ai].s);
		if (as[0] == 0)
			break;
		af = 0;
		for (ai = 0; ai < anum; ai++)
			if (strcmp (as, al[ai].s) == 0)
			{
				af |= al[ai].fl;
				al[ai].fl |= FL_MARK;
			}

		sl = strlen (as);
		asend = strchr (as, 0);
		for (i = 0; i < vnum; i++)
		{
			if ((vc[i] & af) == 0)
				continue;
			if (!anal_match (asend, vl[i]->s))
				continue;

			for (ai = 0; ai < anum; ai++)
			{
				if ((al[ai].fl & FL_MARK) == 0)
					continue;
				if ((al[ai].fl & vc[i]) == 0)
					continue;

				nf = 0;
				switch ((((short)vl[i]->tc[0]) << 8) + (short)vl[i]->tc[1])
				{
				case 0x6E20:	/* 'n ' */
					nf = FL_NOUN;
					break;
				case 0x762D:	/* 'v-' */
					if (strlen (vl[i]->s) >= sl)
						continue;
					nf = FL_VERB | FL_VSUF;
					break;
				case 0x6E73:	/* 'ns' */
					if (strlen (vl[i]->s) >= sl)
						continue;
					nf = FL_NOUN | FL_NSUF | FL_PRO | FL_NAME;
					if (vl[i]->tc[2] == '5' && al[ai].inum == 0)
						nf |= FL_VSUF | FL_VERB | FL_NUM | FL_NUSUF;
					break;
				case 0x7620:	/* 'v ' */
					nf = FL_VPRE;
					for (j = 0; j < al[ai].inum; j++)
					{
						if (vl[al[ai].idx[j]]->tc[0] == 'v' && vl[al[ai].idx[j]]->tc[1] == '-' &&
						    vl[al[ai].idx[j]]->tc[2] == '>' && vl[al[ai].idx[j]]->tc[3] == 'n')
							nf = FL_NOUN;
						else if (vl[al[ai].idx[j]]->tc[0] == 'n' && vl[al[ai].idx[j]]->tc[1] == 's' && vl[al[ai].idx[j]]->tc[2] == '5')
							nf &= ~FL_VPRE;
					}
					break;
				case 0x7673:	/* 'vs' */
					if (strlen (vl[i]->s) >= sl)
						continue;
					nf = FL_VERB | FL_VSUF | FL_PRO;
					for (j = 0; j < al[ai].inum; j++)
						if ((vl[al[ai].idx[j]]->tc[0] == 'v' && vl[al[ai].idx[j]]->tc[1] == '-' &&
						     vl[al[ai].idx[j]]->tc[2] == '>' && vl[al[ai].idx[j]]->tc[3] == 'n') ||
						    (vl[al[ai].idx[j]]->tc[0] == 'n' && vl[al[ai].idx[j]]->tc[1] == 's' && vl[al[ai].idx[j]]->tc[2] == '5'))
							nf &= ~FL_PRO;
					break;
				case 0x6E75:	/* 'nu' */
					if (vl[i]->tc[3] == 's')
					{
						if (strlen (vl[i]->s) >= sl)
							continue;
						nf = FL_NUM | FL_NUSUF;
					}
					else
					{
						if (strlen (vl[i]->s) < sl)
							continue;
					}
					break;
				default:
					if (strlen (vl[i]->s) < sl)
						continue;
					break;
				}

				if (anum >= MAX_ANUM)
				{
					fputs ("mu' HaqwI': internal overflow error\n", stderr);
					continue;
				}
				memcpy (al + anum, al + ai, sizeof(ANAL_REC));
				al[anum].fl = nf;
				al[anum].s[strlen(as) - strlen(vl[i]->s)] = 0;
				al[anum].idx[al[anum].inum++] = i;
				anum++;
			}
		}

		for (ai = 0; ai < anum; ai++)
		{
			if ((al[ai].fl & FL_MARK) == 0)
				continue;
			memcpy (al + ai, al + ai + 1, sizeof(ANAL_REC) * (anum - ai - 1));
			anum--;
			ai--;
		}
	}
}

/*
**	matching routine for the analyzer
*/

short anal_match (char *s, char *m)
{
	char	cs, *ms = m;

	m = strchr (ms, 0);
	do
	{
		if ((cs = *--s) == 0)
			return FALSE;
		if (cs != *--m)
			return FALSE;
	} while (m > ms);
	return TRUE;
}

/*
**	entry point for the dictionary
*/

void dictionary (char *input)
{
	short	i, pcnt;
	char	buf[256], *p, *q, *d, c;
	unsigned char	*up;
	short	mark_for_quit;

	for (p = input; *p != 0; p++)
	{
		if (*p == '`')
			*p = '\'';
	}

	up = (unsigned char *)input;
	while (*up != 0)
		*up++ = upcase[*up];

	for (i = 0; i < vnum; i++)
		vf[i] = 0;

	p = input;
	pcnt = 0;
	mark_for_quit = FALSE;
	while (!mark_for_quit)
	{
		for (; (c = *p) != 0; p++)
			if (c != 32 && c != 9 && c != 13)
				break;
		if (c == 0)
			break;

		for (d = buf; (c = *p) != 0; p++)
		{
			if (c == 32 || c == 9 || c == 13)
				break;
			*d++ = c;
		}
		if (c == 0)
			mark_for_quit = TRUE;
		*d = 0;

		for (i = 0; i < vnum; i++)
		{
			if (vf[i])
				continue;
			if (dict_match (vl[i]->s, buf))
				vf[i] = 1;
			else
			{
				q = strchr (vl[i]->s, 0) + 1;
				if (dict_match (q, buf))
					vf[i] = 1;
			}
		}

		pcnt++;
	}

	if (pcnt == 0)
		for (i = 0; i < vnum; i++)
			vf[i] = 1;

	for (i = 0; i < vnum; i++)
		if (vf[i])
		{
			strcpy (buf, "| ");
			p = vl[i]->s;
			strcat (buf, p);
			p = strchr (p, 0) + 1;
			while (strlen (buf) < 12)
				strcat (buf, " ");
			strcat (buf, " ???? ");
			memcpy (strchr (buf, 0) - 5, vl[i]->tc, 4L);
			strcat (buf, p);
			puts (buf);
		}
}

/*
**	matching routine for the dictionary
*/

short dict_match (char *s, char *pat)
{
	unsigned char	*p1, *p2;

	for (; *s != 0; s++)
	{
		p1 = (unsigned char *)s;
		p2 = (unsigned char *)pat;
		for (;;)
		{
			if (upcase[*p1++] != *p2++)
				break;
			if (*p2 == 0)
				return TRUE;
		}
	}

	return FALSE;
}

/*
**	read in mutay.txt (vocabulary)
*/

void read_mutay (void)
{
	FILE	*f;
	char	buf[256], *p, *p1, *p2, *d;
	short	i, c, class;
	VOCAB_REC	*v;

	puts ("Reading vocabulary...");

	f = fopen ("mutay.txt", "r");
	if (f == NULL)
	{
		fputs ("mu' HaqwI': startup failed (mutay.txt missing)\n", stderr);
		exit (1);
	}

	while (!feof (f) && vnum < MAX_IDX)
	{
		do {
			c = fgetc (f);
		} while (c == 13 || c == 10);
		p = buf;
		while (p < buf + 256 && c != 13 && c != 10 && c != EOF)
		{
			*p++ = c;
			c = fgetc (f);
		}
		while (c != 13 && c != 10 && c != EOF)
			c = fgetc (f);
		*p = 0;

		if (buf[0] == 0 || buf[0] == '#')
			continue;
		if ((v = malloc (sizeof(VOCAB_REC)+strlen(buf))) == NULL)
		{
			fclose (f);
			fputs ("mu' HaqwI': startup failed (out of memory)\n", stderr);
			exit (1);
		}
		strcpy (v->s, buf);
		p1 = strchr (v->s, '=');
		p2 = strchr (v->s, '$');
		if (p1 == NULL || p2 == NULL)
		{
			free (v);
			continue;
		}
		*p1++ = 0;
		*p2++ = 0;
		p = p1;
		d = v->tc;
		while (*p != 0 && d < v->s)
			*d++ = *p++;
		while (d < v->s)
			*d++ = ' ';
		strcpy (p1, p2);
		vl[vnum++] = v;
	}
	if (vnum >= MAX_IDX)
		fputs ("mu' HaqwI': maximum number of words reached - skipped some\n", stderr);

	fclose (f);

	qsort (vl, vnum, sizeof (VOCAB_REC *), voc_sort);

	for (i = 0; i < vnum; i++)
	{
		class = FL_OTHER;
		if (vl[i]->tc[0] == 'n')
		{
			if (vl[i]->tc[1] == ' ')
				class = FL_NOUN;
			else if (vl[i]->tc[1] == 's')
				class = FL_NSUF;
			else if (vl[i]->tc[1] == 'a' && vl[i]->tc[2] == 'm' && vl[i]->tc[3] == 'e')
				class = FL_NAME;
			else if (vl[i]->tc[1] == 'u' && vl[i]->tc[2] == 'm')
			{
				if (vl[i]->tc[3] == ' ')
					class = FL_NUM;
				else if (vl[i]->tc[3] == 's')
					class = FL_NUSUF;
			}
		}
		else if (vl[i]->tc[0] == 'v')
		{
			if (vl[i]->tc[1] == ' ')
				class = FL_VERB;
			else if (vl[i]->tc[1] == 'p')
				class = FL_VPRE;
			else if (vl[i]->tc[1] == 's')
				class = FL_VSUF;
			else if (vl[i]->tc[1] == '-' && vl[i]->tc[2] == '>' && vl[i]->tc[3] == 'n')
				class = FL_NOUN;
		}
		else if (vl[i]->tc[0] == 'p' && vl[i]->tc[1] == 'r' && vl[i]->tc[2] == 'o')
			class = FL_PRO;
		vc[i] = class;
	}
}

/*
**	comparison routine for vocabulary sorting
*/

int voc_sort (const void *v1, const void *v2)
{
	char	*p1, *p2, *f, c1, c2;
	short	l1, l2;

	p1 = (*((VOCAB_REC **)v1))->s;
	p2 = (*((VOCAB_REC **)v2))->s;
	for (;;)
	{
		c1 = *p1++;
		c2 = *p2++;
		if (c1 == 0 || c2 == 0)
			break;
		if ((f = strchr (alph, c1)) == NULL)
			l1 = 500;
		else
			l1 = (short)(f - alph);
		if (c1 == 'c' || c1 == 'g')
			p1++;
		if (c1 == 'n' && *p1 == 'g')
		{
			l1++;
			p1++;
		}
		if (c1 == 't' && *p1 == 'l')
		{
			l1++;
			p1++;
			if (*p1 == 'h')
				p1++;
		}
		if ((f = strchr (alph, c2)) == NULL)
			l2 = 500;
		else
			l2 = (short)(f - alph);
		if (c2 == 'c' || c2 == 'g')
			p2++;
		if (c2 == 'n' && *p2 == 'g')
		{
			l2++;
			p2++;
		}
		if (c2 == 't' && *p2 == 'l')
		{
			l2++;
			p2++;
			if (*p2 == 'h')
				p2++;
		}
		if (l1 == l2)
			continue;
		return l1 - l2;
	}
	if (c1 == 0 && c2 != 0)
		return -1;
	if (c1 != 0 && c2 == 0)
		return 1;
	return 0;
}
